package pds.util;

import java.util.*;
import java.text.*;

/**
 * PPIUtil is a class that contains methods which provide utility services
 * which are useful in other classes. These include conversion utilities,
 * find and replace utilities, etc.
 *
 * @author      Todd King
 * @author      Planetary Data System
 * @version     1.0, 09/16/05
 * @since       1.0
 */
public class PPIUtil {
    /** 
     * Converts a string to a int.
     *
	 * @param value		a string representation of a value.
	 *
	 * @return			the converted value.
	 *
     * @since           1.0
     */
	static public int toInt(String value) {
		return Integer.parseInt(value);
	}
	
    /** 
     * Converts a string to a double.
     *
	 * @param value		a string representation of a value.
	 *
	 * @return			the converted value.
	 *
     * @since           1.0
     */
	static public double toDouble(String value) {
		return Double.parseDouble(value);
	}
	
    /** 
     * Converts a string to a PPITime.
     *
	 * @param pattern	a string representation the pattern the value is
	 *					represented in.
	 * @param value		a string representation of a value.
	 *
	 * @return			the converted value, <code>null</code> if there
	 *					was an error in parsing the value.
	 *
     * @since           1.0
     */
	static public PPITime toTime(String pattern, String value) {
		PPITime	time = new PPITime();
		
		if(!time.convert(pattern, value)) return null;
		
		return time;
	}
	
    /** 
     * Converts a string to a boolean (true/false).
     *
	 * @param value		a string representation of a value.
	 *
	 * @return			if the string is blank, "Y", "YES",
	 *					"T", or "TRUE" then <code>true</code> is returned.
	 *					otherwise <code>false</code> is returned.
	 *					Case is always ignored.
	 *
     * @since           1.0
     */
	static public boolean toBoolean(String value) {
		if(value == null) return false;
		if(value.length() == 0) return true;
		if(value.compareToIgnoreCase("Y") == 0) return true;
		if(value.compareToIgnoreCase("YES") == 0) return true;
		if(value.compareToIgnoreCase("T") == 0) return true;
		if(value.compareToIgnoreCase("TRUE") == 0) return true;
		
		return false;
	}

    /** 
     * Extracts the base name of a file name. The base name is the portion of 
     * the filename without the extension. THe extension if the text following
     * the last period (.) in the filename.
     *
	 * @param filename		the name of the file.
	 *
	 * @return			the basename portion of the file name.
	 *
     * @since           1.0
     */
	static public String basename(String filename) {
		String	basename;
		
		int n = filename.lastIndexOf('.');
		if(n == -1) n = filename.length();
		basename = filename.substring(0, n);
		
		return basename;
	}
	
    /** 
     * Extracts the file name of a path name. The file name is the portion of 
     * the path name which follows the last directory delimiter.
     *
	 * @param pathname		the name of the file including path.
	 *
	 * @return			the file name portion of the path name.
	 *
     * @since           1.0
     */
	static public String filename(String pathname) {
		String	filename;
		
		if(pathname.length() == 0) return "";
		
		int n = pathname.lastIndexOf('/');	// Try Linux style
		if(n == -1) n = pathname.lastIndexOf('\\');	// Try Windows style
		if(n == -1) n = pathname.length();
		filename = pathname.substring(n+1);
		
		return filename;
	}
	
    /** 
     * Extracts the parent directory of a path name. The parent directory is the portion of 
     * the path name up to the last directory delimiter.
     *
	 * @param pathname		the name of the file including path.
	 *
	 * @return			the directory portion of the path name.
	 *
     * @since           1.0
     */
	static public String parent(String pathname) {
		String	parent = "";
		
		int n = pathname.lastIndexOf('/');
		if(n == -1) n = pathname.lastIndexOf('\\');	// Try Windows style
		if(n != -1) parent = pathname.substring(0, n);
		
		return parent;
	}
	
    /** 
     * Extracts the extension of a file name. The extension if the text following
     * the last period (.) in the filename. The period is not included in 
     * returned extension string.
     *
	 * @param filename		the name of the file.
	 *
	 * @return			the extension portion of the file name.
	 *
     * @since           1.0
     */
	static public String extension(String filename) {
		String	extension = "";
		
		int n = filename.lastIndexOf('.');
		if(n != -1) extension = filename.substring(n+1);
		
		return extension;
	}
	
}	
